/*
 * pi - P user Interface
 * 
 * $Id: logic.c,v 1.1 2001/02/05 15:51:07 malekith Exp $
 * Author: Michal Moskal <malekith@pld.org.pl>
 * include COPYING-GNU
 */

#include <pi/pi.h>
#include <unistd.h>

struct pi_object *pi_root_object;

static void focus_move(struct pi_object *ctx, int delta)
{
	struct pi_object *was;

	if (ctx->children == 0)
		return;

	was = ctx->active;

	do {
		if (delta > 0) {
			if (ctx->active) 
				ctx->active = ctx->active->next;
			if (ctx->active == 0)
				ctx->active = ctx->children;
		} else {
			if (ctx->active) 
				ctx->active = ctx->active->prev;
			if (ctx->active == 0) 
				ctx->active = ctx->last_child;
		}
	} while ((ctx->active->flags & pi_f_can_get_focus) == 0 && 
	         ctx->active != was);

	pi_draw(ctx);
}

/*
 * key is handled like this:
 * PRE: root -> window -> button ...
 * NORMAL: ... button -> window -> root
 */
int pi_key_pressed(struct pi_object *ctx, int key)
{
	int (*handler)(struct pi_object*, int);

	if (ctx == 0)
		return 0;

	handler = pi_ptr(prekey_handler);
	if (handler && handler(ctx, key))
		return 1;

	if (pi_key_pressed(ctx->active, key))
		return 1;
		
	handler = pi_ptr(key_handler);
	if (handler && handler(ctx, key))
		return 1;
	return 0;
}

int pi_object_key(struct pi_object *ctx, int key)
{
	if (ctx->active)
		switch (key) {
		case pi_key_right:
		case pi_key_down:
		case '\n':
		case '\t':
			focus_move(ctx, 1);
			return 1;
		case pi_key_left:
		case pi_key_up:
			focus_move(ctx, -1);
			return 1;
		}

	return 0;
}

/* -------------------->  add/kill child */
struct pi_object *pi_add_child(struct pi_object *ctx, struct pi_object *o)
{
	if (ctx->children == 0)
		ctx->last_child = ctx->children = o;
	else {
		ctx->last_child->next = o;
		o->prev = ctx->last_child;
		ctx->last_child = o;
	}

	if (ctx->active == 0 && o->flags & pi_f_can_get_focus)
		ctx->active = o;
		
	o->parent = ctx;

	return o;
}

void pi_kill(struct pi_object *ctx)
{
	void (*kill)(struct pi_object *);
	
	while (ctx->children) 
		pi_kill(ctx->children);

	if (ctx->prev)
		ctx->prev->next = ctx->next;
	if (ctx->next)
		ctx->next->prev = ctx->prev;

	if (ctx == ctx->parent->children) 
		ctx->parent->children = ctx->next;
	if (ctx == ctx->parent->last_child) 
		ctx->parent->last_child = ctx->prev;

	if (ctx->parent->active == ctx)
		ctx->parent->active = 0;
		
	/* ok, we unlinked :) now commit suicide*/
	kill = pi_ptr(kill_handler);

	if (kill)
		kill(ctx);

	pi_kill_object(ctx);
}

/* ---------------> globals */
void pi_add_window(struct pi_object *p)
{
	pi_add_child(pi_root_object, p);
}

int pi_push_event()
{
	int k;
	
	k = pi_getch();
	if (k == 0) {
		usleep(30000);
		return -1;
	}

	k = pi_key_pressed(pi_root_object, k);
	pi_refresh();

	return k;
}

int pi_exec(struct pi_object *win)
{
	struct pi_object *old, *ctx;
	
	old = pi_root_object->active;
	
	if (win->parent == 0)
		pi_add_window(win);

	pi_root_object->active = (struct pi_object*)win;
	
	pi_draw(pi_root_object);
	pi_refresh();

	ctx = win;
	
	while (pi_int(modal_result) == 0)
		pi_push_event();

	return pi_int(modal_result);
}

void pi_repaint()
{
	pi_cls();
	pi_draw(pi_root_object);
	pi_refresh();
}

struct pi_object *pi_new_object(struct pi_object *ctx, struct pi_object *super)
{
	struct pi_object *ret;
	static struct pi_object *super_object;

	if (super_object == 0) 
		super_object = pi_add_object(0,
			pi_tag_flags, 0,
			pi_tag_key_handler, pi_object_key,
			PI_COLOR(norm, lightgray, black),
			PI_COLOR(norm_hl, white, black),
			PI_COLOR(acct, black, green),
			PI_COLOR(acct_hl, lightgreen, green),
			PI_COLOR(spec, black, cyan),
			PI_COLOR(spec_hl, lightcyan, cyan),
		0);

	if (super) {
		pi_set_tag(super, pi_tag_super, super_object);
		ret = pi_add_object(0, pi_tag_super, super, 0);
	} else {
		ret = pi_add_object(0, pi_tag_super, super_object, 0);
		super = super_object;
	}
	
	ret->x = pi_intd(x, -1);
	ret->y = pi_intd(y, -1);
	ret->w = pi_intd(w, -1);
	ret->h = pi_intd(h, -1);
	
	if (pi_intd(flags, -1) != -1)
		ret->flags = pi_int(flags);
	else
		ret->flags = (long)pi_fetch_data(super, pi_tag_flags, 0);

	return ret;
}
